module DC where

import Data.Monoid
import Data.Foldable hiding (toList)
import Data.Maybe

import Prelude hiding (Left, Right)
import Numeric.LinearAlgebra.HMatrix hiding (inv)

type Point = (Double, Double)
type Shape = Point -> Double

circle :: Point -> Double -> Shape
circle (x0, y0) r (x, y) = sqrt ((x0 - x)**2 + (y0 - y)**2) - r

left :: Double -> Shape
left x0 (x, _) = x - x0

right :: Double -> Shape
right x0 (x, _) = x0 - x

lower :: Double -> Shape
lower y0 (_, y) = y - y0

upper :: Double -> Shape
upper y0 (_, y) = y0 - y

(∪) :: Shape -> Shape -> Shape
a ∪ b = \p -> min (a p) (b p)

(∩) :: Shape -> Shape -> Shape
a ∩ b  = \p -> max (a p) (b p)

inv :: Shape -> Shape
inv a = \p -> -(a p)

type Min = Point
type Max = Point

rectangle :: Min -> Max -> Shape
rectangle (xmin, ymin) (xmax, ymax) =
    right xmin ∩ left xmax ∩ upper ymin ∩ lower ymax

hi :: Shape
hi = h ∪ i where
    h = (rectangle (0.1, 0.1) (0.25, 0.9) ∪
         rectangle (0.1, 0.1) (0.6, 0.35) ∪
         circle (0.35, 0.35) 0.25) ∩ inv
         (circle (0.35, 0.35) 0.1 ∪
          rectangle (0.25, 0.1) (0.45, 0.35))
    i = rectangle (0.75, 0.1) (0.9, 0.55) ∪
        circle (0.825, 0.75) 0.1

--------------------------------------------------------------------------------

-- General rule of quadtree ordering:
--   2-------3
--   | 2 | 3 |
--   ---------
--   | 0 | 1 |
--   0-------1


data Tree_ a = Root (Tree_ a) (Tree_ a) (Tree_ a) (Tree_ a) |
               Empty | Full | Leaf a
     deriving (Show, Eq)

type Cell = (Min, Max)
type Tree = Tree_ Cell

buildTree :: Min -> Max -> Int -> Tree
buildTree min max 0 = Leaf (min, max)
buildTree (xmin, ymin) (xmax, ymax) i =
    Root (buildTree (xmin, ymin) (xmid, ymid) (i - 1))
         (buildTree (xmid, ymin) (xmax, ymid) (i - 1))
         (buildTree (xmin, ymid) (xmid, ymax) (i - 1))
         (buildTree (xmid, ymid) (xmax, ymax) (i - 1))
    where xmid = (xmin + xmax) / 2
          ymid = (ymin + ymax) / 2

collapse :: Shape -> Tree -> Tree

collapse shape leaf@(Leaf ((xmin, ymin), (xmax, ymax))) =
    if      all (< 0)  values then  Full
    else if all (>= 0) values then  Empty
    else                            leaf
    where values = [shape (x,y) | x <- [xmin, xmax],
                                  y <- [ymin, ymax]]

collapse shape (Root a b c d) =
    collapse' $ map (collapse shape) [a, b, c, d]
    where collapse' [Empty, Empty, Empty, Empty] = Empty
          collapse' [Full, Full, Full, Full] = Full
          collapse' [q, r, s, t] = Root q r s t

collapse _ t = t

--------------------------------------------------------------------------------

instance Foldable Tree_ where
    foldMap f (Leaf a) = f a
    foldMap f (Root a b c d) = mconcat $ map (foldMap f) [a, b, c, d]
    foldMap _ _ = mempty

--------------------------------------------------------------------------------

-- foldTree tree quad root empty full leaf
foldTree :: Monoid m => Tree -> Cell -> (Cell -> m) -> (Cell -> m)
                                     -> (Cell -> m) -> (Cell -> m) -> m
foldTree (Root a b c d) q@((xmin, ymin), (xmax, ymax)) root empty full leaf =
    mconcat [root q, foldTree a ((xmin, ymin), (xmid, ymid)) root empty full leaf,
                     foldTree b ((xmid, ymin), (xmax, ymid)) root empty full leaf,
                     foldTree c ((xmin, ymid), (xmid, ymax)) root empty full leaf,
                     foldTree d ((xmid, ymid), (xmax, ymax)) root empty full leaf]
    where xmid = (xmin + xmax) / 2
          ymid = (ymin + ymax) / 2

foldTree Empty q _ f _ _ = f q
foldTree Full  q _ _ f _ = f q
foldTree (Leaf _) q _ _ _ f = f q

--------------------------------------------------------------------------------

data Side = Upper | Lower | Left | Right deriving Show

-- This lookup table takes a bitmask abcd and
-- returns a list of edges between which we
-- should draw contours (to outline the shape)
lut :: [[(Side, Side)]]
lut = [[],                          -- 0000
       [(Upper,Right)],             -- 000d
       [(Left,Upper)],              -- 00c0
       [(Left,Right)],              -- 00cd
       [(Right,Lower)],             -- 0b00
       [(Upper,Lower)],             -- 0b0d
       [(Right,Lower),(Left,Upper)],-- 0bc0
       [(Left,Lower)],              -- 0bcd
       [(Lower,Left)],              -- a000
       [(Lower,Left),(Upper,Right)],-- a00d
       [(Lower,Upper)],             -- a0c0
       [(Lower,Right)],             -- a0cd
       [(Right,Left)],              -- ab00
       [(Upper,Left)],              -- ab0d
       [(Right,Upper)],             -- abc0
       []]                          -- abcd

index :: Shape -> Cell -> Int
index shape ((xmin, ymin), (xmax, ymax)) =
    sum [if shape pt < 0 then 2^(3 - i) else 0 |
         (pt, i) <- zip pts [0..]]
    where pts = [(x,y) | y <- [ymin, ymax], x <- [xmin, xmax]]

edges :: Shape -> Cell -> [(Side, Side)]
edges shape c = lut !! index shape c

pt :: Shape -> Cell -> Side -> Point
pt shape ((xmin, ymin), (xmax, ymax)) side =
    case side of Left  -> zero shape (xmin, ymin) (xmin, ymax)
                 Right -> zero shape (xmax, ymin) (xmax, ymax)
                 Lower -> zero shape (xmin, ymin) (xmax, ymin)
                 Upper -> zero shape (xmin, ymax) (xmax, ymax)

zero :: Shape -> Point -> Point -> Point
zero s a@(ax, ay) b@(bx, by)
    | s a >= 0 = zero s b a
    | otherwise = zero' 0.5 0.25 10
    where pos f = (ax * (1-f) + bx * f, ay * (1-f) + by * f)
          zero' f step i =
            if i == 0 then pos f
            else if s (pos f) < 0 then
                    zero' (f + step) (step / 2) (i - 1)
            else    zero' (f - step) (step / 2) (i - 1)

type Edge = (Point, Point)

contours :: Shape -> Cell -> [Edge]
contours shape cell = [(pt' a, pt' b) |
                       (a, b) <- edges shape cell]
    where pt' = pt shape cell

--------------------------------------------------------------------------------
merge :: Shape -> Tree -> Tree
merge shape (Root a b c d) =
    merge' a' b' c' d'
    where [a', b', c', d'] = map (merge shape) [a, b, c, d]
          merge' (Leaf (min, i)) (Leaf (q, r))
                 (Leaf (s, t)) (Leaf (_, max)) =
            let scores = map (score shape (min, max))
                         [i, q, r, s, t]
            in if all (< 0.001) scores
               then Leaf (min, max)
               else Root a' b' c' d'
          merge' _ _ _ _ = Root a' b' c' d'
merge _ t = t

interpolate :: Shape -> Cell -> Point -> Double
interpolate shape ((xmin, ymin), (xmax, ymax)) (x,y) =
    let dx = (x - xmin) / (xmax - xmin)
        dy = (y - ymin) / (ymax - ymin)
        ab = (shape (xmin, ymin)) * (1 - dx) +
             (shape (xmax, ymin)) * dx
        cd = (shape (xmin, ymax)) * (1 - dx) +
             (shape (xmax, ymax)) * dx
    in ab * (1 - dy) + cd * dy

score :: Shape -> Cell -> Point -> Double
score shape cell pt = abs $ (interpolate shape cell pt) - shape pt

--------------------------------------------------------------------------------

deriv :: Shape -> Point -> Point
deriv shape (x,y) =
    let epsilon = 0.001
        dx = shape (x + epsilon, y) - shape (x - epsilon, y)
        dy = shape (x, y + epsilon) - shape (x, y - epsilon)
        len = sqrt $ dx**2 + dy**2
    in (dx / len, dy / len)

feature :: Shape -> Cell -> Maybe Point
feature shape cell =
    if length pts_ >= 2 then
        let pts = map fromTuple pts_
            nms = map (fromTuple . deriv shape) pts_
            center = sum pts / (fromIntegral $ length pts)

            a = fromRows nms
            b = col $ zipWith (\pt nm -> (pt - center) <·> nm) pts nms

            p = center + (toColumns $ linearSolveSVD a b) !! 0
        in Just $ (\[x,y] -> (x,y)) $ (toList p)
    else Nothing
    where pts_ = concatMap (\(a,b) -> [a,b]) $ contours shape cell
          fromTuple = \(x,y) -> fromList [x,y]

--------------------------------------------------------------------------------

dc :: Shape -> Tree -> [Edge]
dc = faceProc

faceProc :: Shape -> Tree -> [Edge]
faceProc shape (Root a b c d) =
    concatMap (faceProc shape) [a,b,c,d] ++
    edgeProcH shape a b ++ edgeProcH shape c d ++
    edgeProcV shape a c ++ edgeProcV shape b d
faceProc _ _ = []

edgeProcH :: Shape -> Tree -> Tree -> [Edge]
edgeProcH shape (Leaf a) (Leaf b) = [(fromJust $ feature shape a,
                                      fromJust $ feature shape b)]
edgeProcH shape leaf@(Leaf _) (Root a _ c _) =
    edgeProcH shape leaf a ++ edgeProcH shape leaf c
edgeProcH shape (Root _ b _ d) leaf@(Leaf _) =
    edgeProcH shape b leaf ++ edgeProcH shape d leaf
edgeProcH shape (Root _ b _ d) (Root a _ c _) =
    edgeProcH shape b a ++ edgeProcH shape d c
edgeProcH _ _ _ = []

edgeProcV :: Shape -> Tree -> Tree -> [Edge]
edgeProcV shape (Leaf a) (Leaf b) = [(fromJust $ feature shape a,
                                      fromJust $ feature shape b)]
edgeProcV shape (Root _ _ c d) leaf@(Leaf _) =
    edgeProcV shape c leaf ++ edgeProcV shape d leaf
edgeProcV shape leaf@(Leaf _) (Root a b _ _) =
    edgeProcV shape leaf a ++ edgeProcV shape leaf b
edgeProcV shape (Root _ _ c d) (Root a b _ _) =
    edgeProcV shape c a ++ edgeProcV shape d b
edgeProcV _ _ _ = []
